/***************************************************************************
 *            server.c
 *
 *  Fri Oct 24 11:28:57 2003
 *  Copyright  2003  Phil Kett
 *  pkett@genesis-midi.com
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
#include "../include/remsync.h"
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <strings.h>
#include <signal.h>

#define BACKLOG 10
int
do_server (int sock)
{
	char message[BIG_BUF], recv_buf[BIG_BUF];
	char version[] = VERSTR;
	int mess_len, count, process_result = 0, pid, loop_count;
	
	/*
	 * Send message to client 
	 */
	sprintf (message, "%s\r\n", version);
	mess_len = strlen (message);
	send (sock, message, mess_len, 0);
	process_result = 0;
	pid = getpid();
	loop_count = 0;
	while (process_result != 2)
	{
		count = read (sock, recv_buf, BIG_BUF);
		if (count != 0)
		{
			process_result = process_recv (recv_buf, sock);
		} else {
			loop_count++;
			if (loop_count > 10) {
				process_result = 2;
			}
		}
		//printf("%d: Waiting for next command\n", pid);
	}
	//printf("Process Result is %d - dropped out of loop\n", process_result);
	return 1;
}

int
start_server (int port)
{
	struct sockaddr_in myaddr;	// server address
	struct sockaddr_in their_addr;
	int sin_size, new_fd, sd = 0;
	char message[BIG_BUF];
	int pid;
	struct sigaction sa;
	int pidfile;
	char pidmsg[BIG_BUF];

	/* daemonize! */
	if (daemon(0,0) < 0)
    {
        perror("daemon");
        exit(2);
    }
	
	if (strlen(conf.pidfile) != 0) {
		/* write out pidfile */
		pidfile = open (conf.pidfile, O_CREAT | O_TRUNC | O_WRONLY | O_BINARY, 0777);
		pid = getpid();
		sprintf(pidmsg, "%d", pid);
		write(pidfile, pidmsg, strlen(pidmsg));
		close(pidfile);	
	}
	sd = socket (PF_INET, SOCK_STREAM, 0);
	//setsockopt (sd, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof (int));
	bzero (&myaddr, sizeof (myaddr));
	myaddr.sin_family = AF_INET;
	myaddr.sin_port = htons (port);
	myaddr.sin_addr.s_addr = INADDR_ANY;
	if (bind (sd, (struct sockaddr *) &myaddr, sizeof (myaddr)) != 0) {
		logmsg("Bind AF_INET");
		exit (-1);
	}
	if (listen (sd, BACKLOG) == -1)
	{
		logmsg("listen");
		exit (1);
	}
	
	sa.sa_handler = sigchld_handler;
	sigemptyset (&sa.sa_mask);
	sa.sa_flags = SA_RESTART;
	if (sigaction (SIGCHLD, &sa, NULL) == -1)
	{
		logmsg("sigaction");
		exit (1);
	}
	
	sa.sa_handler = sighup_handler;
	sigemptyset(&sa.sa_mask);
	sa.sa_flags = SA_RESTART;
	if (sigaction(SIGHUP, &sa, NULL) == -1)
	{
		logmsg("sigaction");
		exit(1);
	}
	
	while (1)
	{
		//printf ("Listening for new connection\n");
		sin_size = sizeof (struct sockaddr_in);
		if ((new_fd =
		     accept (sd, (struct sockaddr *) &their_addr,
			     &sin_size)) == -1)
		{
			logmsg("accept");
			continue;
		}
		
		syslog(LOG_INFO, "Server got connection from %s\n", inet_ntoa (their_addr.sin_addr));
		if (strlen(conf.ip_list) != 0) {
			if (checkip(inet_ntoa(their_addr.sin_addr)) != 1) {
				syslog(LOG_INFO, "Got unauthorised connection from %s\n", inet_ntoa (their_addr.sin_addr));
				sprintf(message, "Permission Denied\r\n");
				send(new_fd, message, strlen(message), 0);
				close(new_fd);
				continue;
			}
		}
		if ((pid=fork()) == 0)
		{
			
			//close(sd); //child process shouldn't do accept()
			if ((pid = fork ()) == 0)
			{		//child process
				//close(sd); //child process shouldn't do accept()
				do_server(new_fd); //actual server process
				//printf("We've come out of the do_server function - should be terminating\n");
				close(new_fd);
				//printf("Next command is the exit(0)\n");
				exit(0);
			} //end of child
			exit(0);
		}
			
	close (new_fd); //parent process doesn't need new_fd
	}
	return 0;
}

int checkip (char *address) {
	char seperator = ',';
	char buffer[BIG_BUF];
	char network_addr[BIG_BUF], mask[BIG_BUF];
	unsigned long int addr1, addr2;
	int len, len2;
	int counter1, counter2;
	int ptr1, ptr2, network = 0, mask_len;
	int shifted1, shifted2, result;
	
	len = strlen(conf.ip_list);
	counter2 = 0;
	for (counter1 = 0; counter1 <= len; counter1++) {
		buffer[counter2] = conf.ip_list[counter1];
		if (conf.ip_list[counter1] == seperator || conf.ip_list[counter1] == '\0') {
			buffer[counter2] = '\0';
			counter2 = 0;
			//printf("Buffer: %s\n", buffer);
			strip_space(buffer);
			//printf("Buffer after strip_space: %s\n", buffer);
			ptr1 = strchr(buffer, '/');
			if (ptr1 != NULL) {
				/* We have a network address so pull out the two parts */
				network = 1;
				ptr2 = &buffer;
				len2 = ptr1 - ptr2;
				strncpy(network_addr, buffer, len2);
				sprintf(mask, "%s", &buffer[len2+1]);
			}
		//printf("Checking %s against %s: Counter: %d\n", address, buffer, counter1);
	/*
	We now have either an IP address in buffer or a network address and mask in
	network_addr and mask
	
	Now to work out whether the address supplies in the function call matches!
	*/
	addr1 = ntohl(inet_addr(address)); /* The supplied IP address */
	
	if (network == 1) {
		/* We're dealing with a network address */
		addr2 = ntohl(inet_addr(network_addr));
		mask_len = 32 - atoi(mask);
		shifted1 = addr1 >> mask_len;
		shifted2 = addr2 >> mask_len;
		result = shifted1 ^ shifted2;
		if (result == 0) {
			return 1;
		} 
	} else {
		/* We're just processing an ip address */
		addr2 = ntohl(inet_addr(buffer));
		if (addr1 == addr2) {
			return 1;
		} 
	}
		} else {
			counter2++;
		}
	}
	
	return 0;
}
